import logging
import datetime as dt

from http import HTTPStatus
from socket import SocketIO
from flask import request, g
from flask_restful import Api

from vantage6.common import logger_name
from vantage6.common.globals import AuthStatus
from vantage6.common.task_status import has_task_finished, TaskStatus
from vantage6.server.resource import ServicesResources, with_user
from vantage6.server import db
from vantage6.server.permission import Scope, Operation, PermissionManager
from vantage6.server.resource.common.input_schema import (
    KillNodeTasksInputSchema,
    KillTaskInputSchema,
)

module_name = logger_name(__name__)
log = logging.getLogger(module_name)


def setup(api: Api, api_base: str, services: dict) -> None:
    """
    Setup the event resource.

    Parameters
    ----------
    api : Api
        Flask restful api instance
    api_base : str
        Base url of the api
    services : dict
        Dictionary with services required for the resource endpoints
    """
    path = "/".join([api_base, module_name])
    log.info(f'Setting up "{path}" and subdirectories')

    api.add_resource(
        KillTask,
        api_base + "/kill/task",
        endpoint="kill_task",
        methods=("POST",),
        resource_class_kwargs=services,
    )

    api.add_resource(
        KillNodeTasks,
        api_base + "/kill/node/tasks",
        endpoint="kill_node_tasks",
        methods=("POST",),
        resource_class_kwargs=services,
    )


# -----------------------------------------------------------------------------
# Permissions
# -----------------------------------------------------------------------------
def permissions(permissions: PermissionManager) -> None:
    """
    Define the permissions for this resource.

    Parameters
    ----------
    permissions : PermissionManager
        Permission manager instance to which permissions are added
    """
    add = permissions.appender(module_name)

    add(
        scope=Scope.ORGANIZATION,
        operation=Operation.RECEIVE,
        description="view websocket events of your organization",
    )
    add(
        scope=Scope.COLLABORATION,
        operation=Operation.RECEIVE,
        description="view websocket events of your collaborations",
    )
    add(
        scope=Scope.GLOBAL,
        operation=Operation.RECEIVE,
        description="view websocket events",
    )
    add(
        scope=Scope.ORGANIZATION,
        operation=Operation.SEND,
        description="send websocket events for your organization",
    )
    add(
        scope=Scope.COLLABORATION,
        operation=Operation.SEND,
        description="send websocket events for your collaborations",
    )
    add(
        scope=Scope.GLOBAL,
        operation=Operation.SEND,
        description="send websocket events to all collaborations",
    )


kill_task_schema = KillTaskInputSchema()
kill_node_tasks_schema = KillNodeTasksInputSchema()


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
class KillTask(ServicesResources):
    """Provide endpoint to kill all containers executing a certain task"""

    def __init__(self, socketio, storage_adapter, mail, api, permissions, config):
        super().__init__(socketio, storage_adapter, mail, api, permissions, config)
        self.r = getattr(self.permissions, module_name)

    @with_user
    def post(self):
        """Kill task specified by task id
        ---
        description: >-
          Kill a task that is currently being executed on a node.

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Event|Global|Create|❌|❌|Create kill signal for any task|\n
          |Event|Collaboration|Create|❌|❌|
          Create kill signal for any task in your collaboration|\n

          Accessible to users.

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  id:
                    type: integer
                    description: Task id which is to be killed

        responses:
          200:
            description: Ok
          404:
            description: Task not found or already completed
          401:
            description: Unauthorized
          400:
            description: No task id provided

        tags: ["Task"]
        """
        body = request.get_json()

        # validate request body
        errors = kill_task_schema.validate(body)
        if errors:
            return {
                "msg": "Request body is incorrect",
                "errors": errors,
            }, HTTPStatus.BAD_REQUEST

        id_ = body.get("id")
        task = db.Task.get(id_)
        if not task:
            return {"msg": f"Task id={id_} not found"}, HTTPStatus.NOT_FOUND

        if has_task_finished(task.status):
            return {
                "msg": f"Task {id_} already finished with status "
                f"'{task.status}', so cannot kill it!"
            }, HTTPStatus.BAD_REQUEST

        # Check permissions. If someone doesn't have global permissions, we
        # check if their organization is part of the appropriate collaboration.
        if not self.r.s_glo.can():
            orgs = task.collaboration.organizations
            if not (self.r.s_col.can() and g.user.organization in orgs):
                return {
                    "msg": "You lack the permission to do that!"
                }, HTTPStatus.UNAUTHORIZED

        # call function to kill the task. This function is outside of the
        # endpoint as it is also used in other endpoints
        kill_task(task, self.socketio)

        return {
            "msg": "Nodes have been instructed to kill any containers running "
            f"for task {id_}."
        }, HTTPStatus.OK


class KillNodeTasks(ServicesResources):
    """Provide endpoint to kill all tasks on a certain node"""

    def __init__(self, socketio, storage_adapter, mail, api, permissions, config):
        super().__init__(socketio, storage_adapter, mail, api, permissions, config)
        self.r = getattr(self.permissions, "task")

    @with_user
    def post(self):
        """Kill all tasks on a node specified by node id
        ---
        description: >-
          Kill all tasks that are currently being executed on a certain node.

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Event|Global|Create|❌|❌|Create kill signal for all tasks on any
          node|\n
          |Event|Collaboration|Create|❌|❌|Create kill signal for all tasks
          on any node in collaborations|\n
          |Event|Organization|Create|❌|❌|Create kill signal for all tasks on
          any node of own organization|\n

          Accessible to users.

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  id:
                    type: integer
                    description: Node id on which tasks are to be killed

        responses:
          200:
            description: Ok
          404:
            description: Node not found
          401:
            description: Unauthorized
          400:
            description: No task id provided or node is not online

        tags: ["Task"]
        """
        body = request.get_json()
        # validate request body
        errors = kill_node_tasks_schema.validate(body)
        if errors:
            return {
                "msg": "Request body is incorrect",
                "errors": errors,
            }, HTTPStatus.BAD_REQUEST

        id_ = body.get("id")
        node = db.Node.get(id_)
        if not node:
            return {"msg": f"Node id={id_} not found"}, HTTPStatus.NOT_FOUND

        if node.status != AuthStatus.ONLINE.value:
            return {
                "msg": f"Node {id_} is not online so cannot kill its tasks!"
            }, HTTPStatus.BAD_REQUEST

        # Check permissions. If someone doesn't have global permissions, we
        # check if their organization is part of the appropriate collaboration.
        if not self.r.s_glo.can():
            collab_orgs = node.collaboration.organizations
            if not (
                (self.r.s_col.can() and g.user.organization in collab_orgs)
                or (
                    self.r.s_org.can()
                    and node.organization_id == g.user.organization_id
                )
            ):
                return {
                    "msg": "You lack the permission to do that!"
                }, HTTPStatus.UNAUTHORIZED

        self.socketio.emit(
            "kill_containers",
            {"node_id": node.id, "collaboration_id": node.collaboration_id},
            namespace="/tasks",
            room=f"collaboration_{node.collaboration_id}",
        )

        return {
            "msg": f"Node {node.id} has been instructed to kill all containers"
            " running on it."
        }, HTTPStatus.OK


def kill_task(task: db.Task, socket: SocketIO) -> None:
    """
    Send instructions to node(s) to kill a certain task

    Parameters
    ----------
    task: Task
        Task that should be killed
    socket: SocketIO
        SocketIO connection object to communicate kill instructions to node
    """
    tasks_to_kill = [task] + list(task.children)
    kill_list = [
        {
            "task_id": task_to_kill.id,
            "run_id": run.id,
            "organization_id": run.organization_id,
        }
        for task_to_kill in tasks_to_kill
        for run in task_to_kill.runs
        if not has_task_finished(run.status)
    ]

    # emit socket event to the node to execute the container kills
    socket.emit(
        "kill_containers",
        {"kill_list": kill_list, "collaboration_id": task.collaboration.id},
        namespace="/tasks",
        room=f"collaboration_{task.collaboration_id}",
    )

    # set tasks and subtasks status to killed
    def set_killed(task: db.Task):
        for run in task.runs:
            if has_task_finished(run.status):
                continue  # don't overwrite status if run is already finished
            run.status = TaskStatus.KILLED
            run.finished_at = dt.datetime.now(dt.timezone.utc)
            run.save()

    set_killed(task)
    for subtask in task.children:
        set_killed(subtask)



import logging
import sqlalchemy.exc

from http import HTTPStatus
from flask import g, request
from flask_restful import Api

from vantage6.common import logger_name
from vantage6.backend.common.resource.error_handling import handle_exceptions
from vantage6.server import db
from vantage6.server.permission import (
    Scope as S,
    Operation as P,
    PermissionManager,
    RuleCollection,
)
from vantage6.server.resource import (
    get_org_ids_from_collabs,
    with_user,
    ServicesResources,
)
from vantage6.server.resource.common.input_schema import UserInputSchema
from vantage6.backend.common.resource.pagination import Pagination
from vantage6.server.resource.common.output_schema import (
    UserSchema,
    UserWithPermissionDetailsSchema,
)


module_name = logger_name(__name__)
log = logging.getLogger(module_name)


def setup(api: Api, api_base: str, services: dict) -> None:
    """
    Setup the user resource.

    Parameters
    ----------
    api : Api
        Flask restful api instance
    api_base : str
        Base url of the api
    services : dict
        Dictionary with services required for the resource endpoints
    """
    path = "/".join([api_base, module_name])
    log.info(f'Setting up "{path}" and subdirectories')

    api.add_resource(
        Users,
        path,
        endpoint="user_without_id",
        methods=("GET", "POST"),
        resource_class_kwargs=services,
    )
    api.add_resource(
        User,
        path + "/<int:id>",
        endpoint="user_with_id",
        methods=("GET", "PATCH", "DELETE"),
        resource_class_kwargs=services,
    )


# ------------------------------------------------------------------------------
# Permissions
# ------------------------------------------------------------------------------
def permissions(permissions: PermissionManager) -> None:
    """
    Define the permissions for this resource.

    Parameters
    ----------
    permissions : PermissionManager
        Permission manager instance to which permissions are added
    """
    add = permissions.appender(module_name)
    add(S.GLOBAL, P.VIEW, description="View any user")
    add(S.COLLABORATION, P.VIEW, description="View users from your collaboration")
    add(S.ORGANIZATION, P.VIEW, description="View users from your organization")
    add(S.GLOBAL, P.CREATE, description="Create a new user for any organization")
    add(
        S.COLLABORATION,
        P.CREATE,
        description="Create a new user for organizations in your " "collaborations",
    )
    add(S.ORGANIZATION, P.CREATE, description="Create a new user for your organization")
    add(S.GLOBAL, P.EDIT, description="Edit any user")
    add(S.COLLABORATION, P.EDIT, description="Edit any user in your collaborations")
    add(S.ORGANIZATION, P.EDIT, description="Edit users from your organization")
    add(S.OWN, P.EDIT, description="Edit your own info")
    add(S.GLOBAL, P.DELETE, description="Delete any user")
    add(S.COLLABORATION, P.DELETE, description="Delete any user in your collaborations")
    add(S.ORGANIZATION, P.DELETE, description="Delete users from your organization")
    add(S.OWN, P.DELETE, description="Delete your own account")


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
user_schema = UserSchema()
user_input_schema = UserInputSchema()
user_schema_with_permissions = UserWithPermissionDetailsSchema()


class UserBase(ServicesResources):
    def __init__(self, socketio, storage_adapter, mail, api, permissions, config):
        super().__init__(socketio, storage_adapter, mail, api, permissions, config)
        self.r: RuleCollection = getattr(self.permissions, module_name)


class Users(UserBase):
    @with_user
    def get(self):
        """List users
        ---
        description: >-
            Returns a list of users that you are allowed to see.

            ### Permission Table\n
            |Rule name|Scope|Operation|Assigned to node|Assigned to container|
            Description|\n
            |--|--|--|--|--|--|\n
            |User|Global|View|❌|❌|View any user details|\n
            |User|Collaboration|View|❌|❌|View user details from your
            collaborations|\n
            |User|Organization|View|❌|❌|View users from your organization|\n

            Accessible to users.

        parameters:
          - in: query
            name: username
            schema:
              type: string
            description: >-
              Name to match with a LIKE operator. \n
              * The percent sign (%) represents zero, one, or multiple
              characters\n
              * underscore sign (_) represents one, single character
          - in: query
            name: organization_id
            schema:
              type: integer
            description: Organization id
          - in: query
            name: collaboration_id
            schema:
              type: integer
            description: Collaboration id
          - in: query
            name: firstname
            schema:
              type: string
            description: >-
              Name to match with a LIKE operator. \n
              * The percent sign (%) represents zero, one, or multiple
              characters\n
              * underscore sign (_) represents one, single character
          - in: query
            name: lastname
            schema:
              type: string
            description: >-
              Name to match with a LIKE operator. \n
              * The percent sign (%) represents zero, one, or multiple
              characters\n
              * underscore sign (_) represents one, single character
          - in: query
            name: email
            schema:
              type: string
            description: >-
              Email to match with a LIKE operator. \n
              * The percent sign (%) represents zero, one, or multiple
              characters\n
              * underscore sign (_) represents one, single character
          - in: query
            name: role_id
            schema:
              type: integer
            description: Role that is assigned to user
          - in: query
            name: rule_id
            schema:
              type: integer
            description: Rule that is assigned to user
          - in: query
            name: last_seen_from
            schema:
              type: date (yyyy-mm-dd)
            description: Show only users seen since this date
          - in: query
            name: last_seen_till
            schema:
              type: date (yyyy-mm-dd)
            description: Show only users last seen before this date
          - in: query
            name: page
            schema:
              type: integer
            description: Page number for pagination (default=1)
          - in: query
            name: per_page
            schema:
              type: integer
            description: Number of items per page (default=10)
          - in: query
            name: sort
            schema:
              type: string
            description: >-
              Sort by one or more fields, separated by a comma. Use a minus
              sign (-) in front of the field to sort in descending order.

        responses:
          200:
            description: Ok
          401:
            description: Unauthorized
          400:
            description: Invalid values provided for request parameters

        security:
            - bearerAuth: []

        tags: ["User"]
        """
        args = request.args
        q = g.session.query(db.User)

        # filter by any field of this endpoint
        for param in ["username", "firstname", "lastname", "email"]:
            if param in args:
                q = q.filter(getattr(db.User, param).like(args[param]))
        if "organization_id" in args:
            if not self.r.allowed_for_org(P.VIEW, args["organization_id"]):
                return {
                    "msg": "You lack the permission view users from the "
                    f'organization with id {args["organization_id"]}!'
                }, HTTPStatus.UNAUTHORIZED
            q = q.filter(db.User.organization_id == args["organization_id"])
        if "last_seen_till" in args:
            q = q.filter(db.User.last_seen <= args["last_seen_till"])
        if "last_seen_from" in args:
            q = q.filter(db.User.last_seen >= args["last_seen_from"])

        # find users with a particulare role or rule assigned
        if "role_id" in args:
            role = db.Role.get(args["role_id"])
            if not role:
                return {
                    "msg": f'Role with id={args["role_id"]} does not exist!'
                }, HTTPStatus.BAD_REQUEST
            # note: We check if role has organization to ensure that users
            # with limited permissions can still see who have default roles
            elif (
                not self.r.allowed_for_org(P.VIEW, role.organization_id)
                and role.organization
            ):
                return {
                    "msg": "You lack the permission view users from the "
                    f"organization that role with id={role.organization_id} "
                    "belongs to!"
                }, HTTPStatus.UNAUTHORIZED
            q = (
                q.join(db.Permission)
                .join(db.Role)
                .filter(db.Role.id == args["role_id"])
            )

        if "rule_id" in args:
            rule = db.Rule.get(args["rule_id"])
            if not rule:
                return {
                    "msg": f'Rule with id={args["rule_id"]} does not exist!'
                }, HTTPStatus.BAD_REQUEST
            q = (
                q.join(db.UserPermission)
                .join(db.Rule)
                .filter(db.Rule.id == args["rule_id"])
            )

        if "collaboration_id" in args:
            if not self.r.can_for_col(P.VIEW, args["collaboration_id"]):
                return {
                    "msg": "You lack the permission view all users from "
                    f'collaboration {args["collaboration_id"]}!'
                }, HTTPStatus.UNAUTHORIZED
            q = q.filter(
                db.User.organization_id.in_(
                    get_org_ids_from_collabs(g.user, args["collaboration_id"])
                )
            )

        # check permissions and apply filter if neccessary
        if not self.r.v_glo.can():
            if self.r.v_col.can():
                q = q.filter(
                    db.User.organization_id.in_(
                        [
                            org.id
                            for col in g.user.organization.collaborations
                            for org in col.organizations
                        ]
                        # include the organization of the user: in case the organization
                        # is not part of any collaboration they would otherwise not be
                        # visible
                        + [g.user.organization_id]
                    )
                )
            elif self.r.v_org.can():
                q = q.filter(db.User.organization_id == g.user.organization_id)
            elif "username" in args and args["username"] == g.user.username:
                # users can always see their own user
                pass
            else:
                return {
                    "msg": "You lack the permission to do that!"
                }, HTTPStatus.UNAUTHORIZED

        # paginate results
        try:
            page = Pagination.from_query(q, request, db.User)
        except (ValueError, AttributeError) as e:
            return {"msg": str(e)}, HTTPStatus.INTERNAL_SERVER_ERROR

        # model serialization
        return self.response(page, user_schema)

    @with_user
    @handle_exceptions
    def post(self):
        """Create user
        ---
        description: >-
          Creates new user from the request data to the users organization.\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |User|Global|Create|❌|❌|Create a new user|\n
          |User|Collaboration|Create|❌|❌|Create a new user for any
          organization in your collaborations|\n
          |User|Organization|Create|❌|❌|Create a new user as part of your
          organization|\n

          Accessible to users.

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  username:
                    type: string
                    description: Unique username
                  firstname:
                    type: string
                    description: First name
                  lastname:
                    type: string
                    description: Last name
                  password:
                    type: string
                    description: Password
                  organization_id:
                    type: integer
                    description: Organization id to which user is assigned
                  roles:
                    type: array
                    items:
                      type: integer
                    description: User's roles
                  rules:
                    type: array
                    items:
                      type: integer
                    description: Extra rules for the user on top of the roles
                  email:
                    type: string
                    description: Email address

        responses:
          201:
            description: Ok
          400:
            description: Username or email already exists
          401:
            description: Unauthorized
          404:
            description: Organization id does not exist

        security:
          - bearerAuth: []

        tags: ["User"]
        """
        data = request.get_json()
        # validate request body
        errors = user_input_schema.validate(data)
        if errors:
            return {
                "msg": "Request body is incorrect",
                "errors": errors,
            }, HTTPStatus.BAD_REQUEST

        # check unique constraints
        if db.User.username_exists(data["username"]):
            return {"msg": "username already exists."}, HTTPStatus.BAD_REQUEST

        if db.User.exists("email", data["email"]):
            return {"msg": "email already exists."}, HTTPStatus.BAD_REQUEST

        # check if the organization has been provided, if this is the case the
        # user needs global permissions in case it is not their own
        organization_id = g.user.organization_id
        if data.get("organization_id"):
            if data["organization_id"] != organization_id:
                if self.r.c_glo.can():
                    # check if organization exists
                    org = db.Organization.get(data["organization_id"])
                    if not org:
                        return {
                            "msg": "Organization does not exist."
                        }, HTTPStatus.NOT_FOUND
            organization_id = data["organization_id"]

        # check that user is allowed to create users
        if not self.r.allowed_for_org(P.CREATE, organization_id):
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        # process the required roles. It is only possible to assign roles with
        # rules that you already have permission to. This way we ensure you can
        # never extend your power on your own.
        potential_roles = data.get("roles")
        roles = []
        if potential_roles:
            for role in potential_roles:
                role_ = db.Role.get(role)
                if role_:
                    self.permissions.check_user_rules(role_.rules)
                    roles.append(role_)

                    # validate that the assigned role is either a general role
                    # or a role pertaining to that organization
                    if role_.organization and role_.organization.id != organization_id:
                        return {
                            "msg": (
                                "You can't assign that role as the role belongs to"
                                " a different organization than the user."
                            )
                        }, HTTPStatus.UNAUTHORIZED

        # You can only assign rules that you already have to others.
        potential_rules = data.get("rules")
        rules = []
        if potential_rules:
            rules = [db.Rule.get(rule) for rule in potential_rules if db.Rule.get(rule)]
            self.permissions.check_user_rules(rules)

        # Ok, looks like we got most of the security hazards out of the way
        user = db.User(
            username=data["username"],
            firstname=data["firstname"],
            lastname=data["lastname"],
            roles=roles,
            rules=rules,
            organization_id=organization_id,
            email=data["email"],
            password=data["password"],
        )

        # check if the password meets password criteria
        msg = user.set_password(data["password"])
        if msg:
            return {"msg": msg}, HTTPStatus.BAD_REQUEST

        user.save()

        return user_schema.dump(user), HTTPStatus.CREATED


class User(UserBase):
    @with_user
    def get(self, id):
        """Get user
        ---
        description: >-
            Returns the user specified by the id.\n

            ### Permission Table\n
            |Rule name|Scope|Operation|Assigned to node|Assigned to container|
            Description|\n
            |-- |--|--|--|--|--|\n
            |User|Global|View|❌|❌|View any user details|\n
            |User|Collaboration|View|❌|❌|View users from your
            collaborations|\n
            |User|Organization|View|❌|❌|View users from your
            organization|\n
            |User|Organization|Own|❌|❌|View details about your own user|\n

            Accessible to users.

        parameters:
            - in: path
              name: id
              schema:
                type: integer
              description: User id
              required: true
            - in: path
              name: include_permissions
              schema:
                type: boolean
              description: Whether or not to include extra permission info for
                the user. By default false.

        responses:
            200:
                description: Ok
            404:
                description: User not found
            401:
                description: Unauthorized

        security:
            - bearerAuth: []

        tags: ["User"]
        """
        user = db.User.get(id)
        if not user:
            return {"msg": f"user id={id} is not found"}, HTTPStatus.NOT_FOUND

        schema = user_schema

        if request.args.get("include_permissions", False):
            schema = user_schema_with_permissions

        # allow user to be returned if authenticated user can view users from
        # that organization or if the user is the same as the authenticated
        # user.
        same_user = g.user.id == user.id
        if same_user or self.r.allowed_for_org(P.VIEW, user.organization_id):
            return schema.dump(user, many=False), HTTPStatus.OK
        else:
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

    @with_user
    @handle_exceptions
    def patch(self, id):
        """Update user
        ---
        description: >-
          Update user information.\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |User|Global|Edit|❌|❌|Edit any user|\n
          |User|Collaboration|Edit|❌|❌|Edit users in your collaborations|\n
          |User|Organization|Edit|❌|❌|Edit users in your organization|\n
          |User|Own|Edit|❌|❌|Edit your own user account|\n

          Accessible to users.

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  username:
                    type: string
                    description: Unique username
                  firstname:
                    type: string
                    description: First name
                  lastname:
                    type: string
                    description: Last name
                  email:
                    type: string
                    description: Email address
                  roles:
                    type: array
                    items:
                      type: integer
                    description: User's roles
                  rules:
                    type: array
                    items:
                      type: integer
                    description: Extra rules for the user on top of the roles

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: User id
            required: true

        responses:
          200:
            description: Ok
          400:
            description: User cannot be updated to contents of request body,
              e.g. due to duplicate email address.
          404:
            description: User not found
          401:
            description: Unauthorized

        security:
          - bearerAuth: []

        tags: ["User"]
        """
        user = db.User.get(id)
        if not user:
            return {"msg": f"user id={id} not found"}, HTTPStatus.NOT_FOUND

        data = request.get_json()
        # validate request body
        errors = user_input_schema.validate(data, partial=True)
        if errors:
            return {
                "msg": "Request body is incorrect",
                "errors": errors,
            }, HTTPStatus.BAD_REQUEST
        if data.get("password"):
            return {
                "msg": "You cannot change your password here!"
            }, HTTPStatus.BAD_REQUEST

        # check permissions
        if not (self.r.e_own.can() and user == g.user) and not self.r.allowed_for_org(
            P.EDIT, user.organization_id
        ):
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        # update user and check for unique constraints
        if data.get("username") is not None:
            if user.username != data["username"]:
                if db.User.exists("username", data["username"]):
                    return {
                        "msg": "User with that username already exists"
                    }, HTTPStatus.BAD_REQUEST
                elif user.id != g.user.id:
                    return {
                        "msg": "You cannot change the username of another user"
                    }, HTTPStatus.BAD_REQUEST
            user.username = data["username"]
        if data.get("firstname") is not None:
            user.firstname = data["firstname"]
        if data.get("lastname") is not None:
            user.lastname = data["lastname"]
        if data.get("email") is not None:
            if user.email != data["email"] and db.User.exists("email", data["email"]):
                return {
                    "msg": "User with that email already exists."
                }, HTTPStatus.BAD_REQUEST
            user.email = data["email"]

        # request parser is awefull with lists
        if "roles" in data:
            # validate that these roles exist
            roles = []
            for role_id in data["roles"]:
                role = db.Role.get(role_id)
                if not role:
                    return {
                        "msg": f"Role={role_id} can not be found!"
                    }, HTTPStatus.NOT_FOUND
                roles.append(role)

            # validate that user is not changing their own roles
            if user == g.user:
                return {
                    "msg": "You can't changes your own roles!"
                }, HTTPStatus.UNAUTHORIZED

            # validate that user can assign these
            for role in roles:
                self.permissions.check_user_rules(role.rules)

                # validate that the assigned role is either a general role or a
                # role pertaining to that organization
                if role.organization and role.organization.id != user.organization_id:
                    return {
                        "msg": (
                            "You can't assign that role to that user as the role "
                            "belongs to a different organization than the user "
                        )
                    }, HTTPStatus.UNAUTHORIZED

            # validate that user is not deleting roles they cannot assign
            # e.g. an organization admin is not allowed to delete a root role
            deleted_roles = [r for r in user.roles if r not in roles]
            for role in deleted_roles:
                self.permissions.check_user_rules(role.rules)
            user.roles = roles

        if "rules" in data:
            # validate that these rules exist
            rules = []
            for rule_id in data["rules"]:
                rule = db.Rule.get(rule_id)
                if not rule:
                    return {
                        "msg": f"Rule={rule_id} can not be found!"
                    }, HTTPStatus.NOT_FOUND
                rules.append(rule)

            # validate that user is not changing their own rules
            if user == g.user:
                return {
                    "msg": "You can't changes your own rules!"
                }, HTTPStatus.UNAUTHORIZED

            # validate that user can assign these
            self.permissions.check_user_rules(rules)

            # validate that user is not deleting rules they do not have
            # themselves
            deleted_rules = [r for r in user.rules if r not in rules]
            self.permissions.check_user_rules(deleted_rules)

            user.rules = rules

        try:
            user.save()
        except sqlalchemy.exc.IntegrityError as e:
            log.error(e)
            user.session.rollback()
            return {
                "msg": "User could not be updated with those parameters."
            }, HTTPStatus.BAD_REQUEST

        return user_schema.dump(user), HTTPStatus.OK

    @with_user
    def delete(self, id):
        """Remove user.
        ---
        description: >-
          Delete a user account permanently.\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |User|Global|Delete|❌|❌|Delete any user|\n
          |User|Collaboration|Delete|❌|❌|Delete users from your
          collaboration|\n
          |User|Organization|Delete|❌|❌|Delete users from your
          organization|\n
          |User|Own|Delete|❌|❌|Delete your own account|\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: User id
            required: true
          - in: query
            name: delete_dependents
            schema:
              type: boolean
            description: If set to true, the user will be deleted along with
              all tasks they created (default=False)

        responses:
          200:
            description: Ok
          404:
            description: User not found
          401:
            description: Unauthorized

        security:
          - bearerAuth: []

        tags: ["User"]
        """
        user = db.User.get(id)
        if not user:
            return {"msg": f"user id={id} not found"}, HTTPStatus.NOT_FOUND

        if not (self.r.d_own.can() and user == g.user) and not self.r.allowed_for_org(
            P.DELETE, user.organization_id
        ):
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        # check if user created any tasks
        if user.created_tasks:
            params = request.args
            if not params.get("delete_dependents", False):
                return {
                    "msg": f"User has created {len(user.created_tasks)} tasks."
                    " Please delete those first, or set the "
                    "`delete_dependents` parameter to true to delete them "
                    "automatically together with this user."
                }, HTTPStatus.BAD_REQUEST
            else:
                log.warn(
                    f"Deleting {len(user.created_tasks)} tasks created by"
                    f" user id={id}"
                )
                for task in user.created_tasks:
                    task.delete()

        user.delete()
        log.info(f"user id={id} is removed from the database")
        return {"msg": f"user id={id} is removed from the database"}, HTTPStatus.OK



import logging

from flask import request, g
from flask_restful import Api
from http import HTTPStatus

from vantage6.server import db
from vantage6.backend.common.resource.pagination import Pagination
from vantage6.server.resource.common.input_schema import (
    StudyChangeOrganizationSchema,
    StudyInputSchema,
)
from vantage6.server.permission import (
    RuleCollection,
    Scope as S,
    Operation as P,
    PermissionManager,
)
from vantage6.server.resource.common.output_schema import (
    OrganizationSchema,
    StudySchema,
    StudyWithOrgsSchema,
)
from vantage6.server.resource import with_user, only_for, ServicesResources


module_name = __name__.split(".")[-1]
log = logging.getLogger(module_name)


def setup(api: Api, api_base: str, services: dict) -> None:
    """
    Setup the study resource.

    Parameters
    ----------
    api : Api
        Flask restful api instance
    api_base : str
        Base url of the api
    services : dict
        Dictionary with services required for the resource endpoints
    """
    path = "/".join([api_base, module_name])
    log.info('Setting up "%s" and subdirectories', path)

    api.add_resource(
        Studies,
        path,
        endpoint="study_without_id",
        methods=("GET", "POST"),
        resource_class_kwargs=services,
    )
    api.add_resource(
        Study,
        path + "/<int:id>",
        endpoint="study_with_id",
        methods=("GET", "PATCH", "DELETE"),
        resource_class_kwargs=services,
    )
    api.add_resource(
        StudyOrganization,
        path + "/<int:id>/organization",
        endpoint="study_with_id_organization",
        methods=("POST", "DELETE"),
        resource_class_kwargs=services,
    )


# # Schemas
study_schema = StudySchema()
org_schema = OrganizationSchema()
study_with_orgs_schema = StudyWithOrgsSchema()
study_input_schema = StudyInputSchema()
study_change_org_schema = StudyChangeOrganizationSchema()


# -----------------------------------------------------------------------------
# Permissions
# -----------------------------------------------------------------------------
def permissions(permissions: PermissionManager) -> None:
    """
    Define the permissions for this resource.

    Parameters
    ----------
    permissions : PermissionManager
        Permission manager instance to which permissions are added
    """

    add = permissions.appender(module_name)

    add(scope=S.GLOBAL, operation=P.VIEW, description="view any study")
    add(
        scope=S.ORGANIZATION,
        operation=P.VIEW,
        assign_to_container=True,
        assign_to_node=True,
        description="view studies that your organization is involved in",
    )
    add(
        scope=S.COLLABORATION,
        operation=P.VIEW,
        description="view studies within your collaborations",
    )
    add(scope=S.GLOBAL, operation=P.EDIT, description="edit any study")
    add(
        scope=S.COLLABORATION,
        operation=P.EDIT,
        description="edit studies within your collaborations",
    )
    add(
        scope=S.ORGANIZATION,
        operation=P.EDIT,
        description="edit studies that your organization is involved in",
    )

    add(scope=S.GLOBAL, operation=P.CREATE, description="create a new study")
    add(
        scope=S.COLLABORATION,
        operation=P.CREATE,
        description="create a new study within your collaborations",
    )

    add(scope=S.GLOBAL, operation=P.DELETE, description="delete a study")
    add(
        scope=S.COLLABORATION,
        operation=P.DELETE,
        description="delete any study within your collaborations",
    )
    add(
        scope=S.ORGANIZATION,
        operation=P.DELETE,
        description="delete studies that your organization is involved in",
    )


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
class StudyBase(ServicesResources):
    def __init__(self, socketio, storage_adapter, mail, api, permissions, config):
        super().__init__(socketio, storage_adapter, mail, api, permissions, config)
        self.r: RuleCollection = getattr(self.permissions, module_name)

    def _select_schema(self) -> StudySchema:
        """
        Select the output schema based on which resources should be included

        Returns
        -------
        StudySchema or derivative of it
            Schema to use for serialization
        """
        if self.is_included("organizations"):
            return study_with_orgs_schema
        else:
            return study_schema

    def _check_membership_collaboration(
        self, organization_ids: list[int], collaboration_id: int
    ) -> bool:
        """
        Check whether organizations are all members of a collaboration

        Parameters
        ----------
        organization_ids : list[int]
            List of organization ids
        collaboration_id : int
            Collaboration id

        Returns
        -------
        bool
            True if all organizations are members of the collaboration, False otherwise
        """
        collab = db.Collaboration.get(collaboration_id)
        return set(organization_ids).issubset(
            set([org.id for org in collab.organizations])
        )


class Studies(StudyBase):
    """Resource for /api/study."""

    @only_for(("user", "node"))
    def get(self):
        """Returns a list of studies, which are subsets of organizations from a
        collaboration
        ---
        description: >-
          Returns a list of studies. Depending on your permission, all
          studies are shown or only studies in which your organization participates.
          See the table bellow.\n\n

          ### Permission Table\n
          |Rulename|Scope|Operation|Assigned to Node|Assigned to Container|
          Description|\n
          | -- | -- | -- | -- | -- | -- |\n
          |Study|Global|View|❌|❌|All studies|\n
          |Study|Collaboration|View|❌|❌|Studies which are part of collaborations
          that your organization participates in|\n
          |Study|Organization|View|✅|✅|Studies in which
          your organization participates |\n\n

          Accessible to users.

        parameters:
          - in: query
            name: name
            schema:
              type: string
            description: >-
              Name to match with a LIKE operator. \n
              * The percent sign (%) represents zero, one, or multiple
              characters\n
              * underscore sign (_) represents one, single character
          - in: query
            name: organization_id
            schema:
              type: integer
            description: Filter studies by organization id
          - in: query
            name: collaboration_id
            schema:
              type: integer
            description: Filter studies by collaboration id
          - in: query
            name: include
            schema:
              type: array
              items:
                type: string
            description: Include 'organizations' to include the organizations
              in the output.
          - in: query
            name: page
            schema:
              type: integer
            description: Page number for pagination (default=1)
          - in: query
            name: per_page
            schema:
              type: integer
            description: Number of items per page (default=10)
          - in: query
            name: sort
            schema:
              type: string
            description: >-
              Sort by one or more fields, separated by a comma. Use a minus
              sign (-) in front of the field to sort in descending order.

        responses:
          200:
            description: Ok
          401:
            description: Unauthorized
          400:
            description: Improper values for pagination or sorting parameters

        security:
          - bearerAuth: []

        tags: ["Collaboration"]
        """

        # obtain organization from authenticated
        auth_org_id = self.obtain_organization_id()
        auth_collab_ids = self.obtain_auth_collaboration_ids()
        q = g.session.query(db.Study)
        args = request.args

        # filter by a field of this endpoint
        if "name" in args:
            q = q.filter(db.Study.name.like(args["name"]))

        # find studies containing a specific organization
        if "organization_id" in args:
            if self.r.v_glo.can() or self.r.v_col.can():
                # find all studies containing the organization. Note that for the
                # collaboration scope, we filter by collaboration later on
                q = (
                    q.join(db.StudyMember)
                    .join(db.Organization)
                    .filter(db.Organization.id == args["organization_id"])
                )
            elif self.r.v_org.can() and args["organization_id"] == str(auth_org_id):
                # the user can already only view filters of their own organization, so
                # argument is superfluous: no additional filters necessary
                pass
            else:
                return {
                    "msg": "You lack the permission to request studies for this "
                    "organization!"
                }, HTTPStatus.UNAUTHORIZED

        # Note that the collaboration_id filter is not allowed in most endpoints if the
        # user only has organization permission - in those cases they will often
        # unknowingly miss part of the resources. Here it is allowed because there is a
        # clear use case: the user wants to see all their studies
        if "collaboration_id" in args:
            if self.r.v_glo.can() or (
                (self.r.v_col.can() or self.r.v_org.can())
                and int(args["collaboration_id"]) in auth_collab_ids
            ):
                q = q.filter(db.Study.collaboration_id == args["collaboration_id"])
            else:
                return {
                    "msg": "You lack the permission to request studies for this "
                    "collaboration!"
                }, HTTPStatus.UNAUTHORIZED

        # filter based on permissions
        if not self.r.v_glo.can():
            if self.r.v_col.can():
                q = q.join(
                    db.Collaboration,
                    db.Study.collaboration_id == db.Collaboration.id,
                ).filter(db.Collaboration.id.in_(auth_collab_ids))
            elif self.r.v_org.can():
                q = q.join(db.Organization, db.Study.organizations).filter(
                    db.Study.organizations.any(id=auth_org_id)
                )
            else:
                return {
                    "msg": "You lack the permission to do that!"
                }, HTTPStatus.UNAUTHORIZED

        # paginate the results
        try:
            page = Pagination.from_query(q, request, db.Study)
        except (ValueError, AttributeError) as e:
            return {"msg": str(e)}, HTTPStatus.INTERNAL_SERVER_ERROR

        schema = self._select_schema()

        # serialize models
        return self.response(page, schema)

    @with_user
    def post(self):
        """Create study
        ---
        description: >-
          Create a new study between organizations.\n\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to Node|Assigned to
          Container|Description|\n
          |--|--|--|--|--|--|\n
          |Study|Global|Create|❌|❌|Create study in any collaboration|\n\n
          |Study|Collaboration|Create|❌|❌|Create study in collaborations that your
          organization participates in|\n\n

          Accessible to users.

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  name:
                    type: string
                    description: Unique human readable name for the study
                  collaboration_id:
                    type: integer
                    description: Collaboration id to which the study belongs
                  organization_ids:
                    type: array
                    items:
                      type: integer
                      description: List of organization ids which form the
                        study

        responses:
          200:
            description: Ok
          400:
            description: Study name already exists, or not all organizations are member
              of the collaboration
          401:
            description: Unauthorized

        security:
          - bearerAuth: []

        tags: ["Collaboration"]
        """
        data = request.get_json()
        # validate request body
        errors = study_input_schema.validate(data)
        if errors:
            return {
                "msg": "Request body is incorrect",
                "errors": errors,
            }, HTTPStatus.BAD_REQUEST

        name = data["name"]
        if db.Study.exists("name", name):
            return {
                "msg": f"Study with name '{name}' already exists!"
            }, HTTPStatus.BAD_REQUEST

        if not self.r.can_for_col(P.CREATE, data["collaboration_id"]):
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        organizations_in_collab = self._check_membership_collaboration(
            data["organization_ids"], data["collaboration_id"]
        )
        if not organizations_in_collab:
            return {
                "msg": "Not all organizations are part of the collaboration!"
            }, HTTPStatus.BAD_REQUEST

        study = db.Study(
            name=name,
            collaboration_id=data["collaboration_id"],
            organizations=[
                db.Organization.get(org_id)
                for org_id in data["organization_ids"]
                if db.Organization.get(org_id)
            ],
        )

        study.save()
        return study_with_orgs_schema.dump(study), HTTPStatus.CREATED


class Study(StudyBase):
    """Resource for /api/study/<int:id>."""

    @only_for(("user", "node", "container"))
    def get(self, id):
        """Get study
        ---
        description: >-
          Returns the study with the specified id.\n

          ### Permission Table\n
          |Rulename|Scope|Operation|Assigned to Node|Assigned to Container|
          Description|\n
          | -- | -- | -- | -- | -- | -- |\n
          |Study|Global|View|❌|❌|All studies|\n
          |Study|Collaboration|View|❌|❌|Studies which are part of collaborations
          that your organization participates in|\n
          |Study|Organization|View|✅|✅|Studies in which
          your organization participates |\n\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
              minimum: 1
            description: Study id
            required: true
          - in: query
            name: include
            schema:
              type: array
              items:
                type: string
            description: Include 'organizations' to include the organizations
              within the study.

        responses:
          200:
            description: Ok
          404:
            description: Study with specified id is not found
          401:
            description: Unauthorized

        security:
          - bearerAuth: []

        tags: ["Collaboration"]
        """
        study = db.Study.get(id)

        # check that study exists
        if not study:
            return {"msg": f"Study with id={id} not found"}, HTTPStatus.NOT_FOUND

        # obtain the organization id of the authenticated
        auth_org_id = self.obtain_organization_id()
        auth_collab_ids = self.obtain_auth_collaboration_ids()

        # verify that the right permissions are present
        org_ids = [org.id for org in study.organizations]
        if (
            not self.r.v_glo.can()
            and not (self.r.v_col.can() and study.collaboration_id in auth_collab_ids)
            and not (self.r.v_org.can() and auth_org_id in org_ids)
        ):
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        schema = self._select_schema()

        return schema.dump(study, many=False), HTTPStatus.OK  # 200

    @with_user
    def patch(self, id):
        """Update study
        ---
        description: >-
          Updates the study with the specified id.\n\n
          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Study|Global|Edit|❌|❌|Update a study|\n
          |Study|Collaboration|Edit|❌|❌|Update a study within collaborations that
          your organization is a member of|\n\n
          |Study|Organization|Edit|❌|❌|Update a study that your organization is a
          member of|\n\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: Study id
            required: tr

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  name:
                    type: string
                    description: Human readable label
                  organization_ids:
                    type: array
                    items:
                      type: integer
                    description: List of organization ids

        responses:
          200:
            description: Ok
          404:
            description: Study with specified id is not found
          401:
            description: Unauthorized
          400:
            description: Study name already exists, organizations are not
              collaboration members, or request body is incorrect

        security:
          - bearerAuth: []

        tags: ["Collaboration"]
        """
        study = db.Study.get(id)

        # check if collaboration exists
        if not study:
            return {
                "msg": f"Study with id={id} can not be found"
            }, HTTPStatus.NOT_FOUND  # 404

        # verify permissions
        auth_org_id = self.obtain_organization_id()
        if (
            not self.r.e_glo.can()
            and not (
                self.r.e_col.can()
                and study.collaboration_id in self.obtain_auth_collaboration_ids()
            )
            and not (
                self.r.e_org.can()
                and auth_org_id in [org.id for org in study.organizations]
            )
        ):
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        data = request.get_json()
        # validate request body
        errors = study_input_schema.validate(data, partial=True)
        if errors:
            return {
                "msg": "Request body is incorrect",
                "errors": errors,
            }, HTTPStatus.BAD_REQUEST

        # only update fields that are provided
        if "name" in data:
            name = data["name"]
            if study.name != name and db.Study.exists("name", name):
                return {
                    "msg": f"Study name '{name}' already exists!"
                }, HTTPStatus.BAD_REQUEST
            study.name = name
        if "organization_ids" in data:
            # check that all organizations are member of the collaboration
            organizations_in_collab = self._check_membership_collaboration(
                data["organization_ids"], study.collaboration_id
            )
            if not organizations_in_collab:
                return {
                    "msg": "Not all organizations are part of the collaboration!"
                }, HTTPStatus.BAD_REQUEST
            study.organizations = [
                db.Organization.get(org_id)
                for org_id in data["organization_ids"]
                if db.Organization.get(org_id)
            ]

        study.save()

        return (
            study_with_orgs_schema.dump(study, many=False),
            HTTPStatus.OK,
        )  # 200

    @with_user
    def delete(self, id):
        """Delete study
        ---
        description: >-
          Removes the study from the database entirely.

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Study|Global|Delete|❌|❌|Remove study|\n
          |Study|Collaboration|Delete|❌|❌|Remove studies from collaborations that
          your organization is a member of|\n\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: Study id
            required: true
          - in: query
            name: delete_dependents
            schema:
              type: boolean
            description: If set to true, the study will be deleted along with all its
              tasks (default=False)

        responses:
          200:
            description: Ok
          404:
            description: Study with specified id is not found
          401:
            description: Unauthorized

        security:
          - bearerAuth: []

        tags: ["Collaboration"]
        """

        study = db.Study.get(id)
        if not study:
            return {"msg": f"Study with id={id} not found"}, HTTPStatus.NOT_FOUND

        # verify permissions
        if (
            not self.r.d_glo.can()
            and not (
                self.r.d_col.can()
                and study.collaboration_id in self.obtain_auth_collaboration_ids()
            )
            and not (
                self.r.d_org.can()
                and self.obtain_organization_id()
                in [org.id for org in study.organizations]
            )
        ):
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        if study.tasks:
            delete_dependents = request.args.get("delete_dependents", False)
            if not delete_dependents:
                return {
                    "msg": f"Study id={id} has {len(study.tasks)} tasks. Please delete"
                    " them separately or set delete_dependents=True"
                }, HTTPStatus.BAD_REQUEST
            else:
                log.warning(
                    "Deleting study id=%s along with %s tasks",
                    id,
                    len(study.tasks),
                )
                for task in study.tasks:
                    task.delete()

        study.delete()
        return {"msg": f"Study id={id} successfully deleted"}, HTTPStatus.OK


class StudyOrganization(ServicesResources):
    """Resource for /api/study/<int:id>/organization."""

    def __init__(self, socketio, storage_adapter, mail, api, permissions, config):
        super().__init__(socketio, storage_adapter, mail, api, permissions, config)
        self.r: RuleCollection = getattr(self.permissions, module_name)

    @with_user
    def post(self, id):
        """Add organization to study
        ---
        description: >-
          Adds a single organization to an existing study.\n\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Study|Global|Edit|❌|❌|Add organization to a study|\n
          |Study|Collaboration|Edit|❌|❌|Add organization to a study that your
          organization is already a member of|\n\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: Study id
            required: tr

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  id:
                    type: integer
                    description: Organization id which needs to be added

        responses:
          200:
            description: Ok
          404:
            description: Specified study or organization does not exist
          401:
            description: Unauthorized

        security:
          - bearerAuth: []

        tags: ["Collaboration"]
        """
        # get collaboration to which te organization should be added
        study = db.Study.get(id)
        if not study:
            return {"msg": f"Study with id={id} can not be found"}, HTTPStatus.NOT_FOUND

        # verify permissions
        if not self.r.can_for_col(P.EDIT, study.collaboration_id):
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        # validate request body
        data = request.get_json()
        errors = study_change_org_schema.validate(data)
        if errors:
            return {
                "msg": "Request body is incorrect",
                "errors": errors,
            }, HTTPStatus.BAD_REQUEST

        # get the organization
        organization = db.Organization.get(data["id"])
        if not organization:
            return {
                "msg": f"organization with id={id} is not found"
            }, HTTPStatus.NOT_FOUND

        # check that organization is collaboration member
        if not organization in study.collaboration.organizations:
            return {
                "msg": "Organization is not a member of the collaboration!"
            }, HTTPStatus.BAD_REQUEST

        # append organization to the collaboration
        study.organizations.append(organization)
        study.save()
        return org_schema.dump(study.organizations, many=True), HTTPStatus.OK

    @with_user
    def delete(self, id):
        """Remove organization from a study
        ---
        description: >-
          Removes a single organization from an existing study.\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Study|Global|Edit|❌|❌|Remove an organization from an
          existing study|\n
          |Study|Collaboration|Edit|❌|❌|Remove an organization from
          an existing study within collaborations that your organization is a member
          of|\n\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: Study id
            required: true

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  id:
                    type: integer
                    description: Organization id which needs to be deleted

        responses:
          200:
            description: Ok
          404:
            description: Specified collaboration or organization does not exist
          401:
            description: Unauthorized

        tags: ["Collaboration"]
        """
        # get collaboration from which organization should be removed
        study = db.Study.get(id)
        if not study:
            return {"msg": f"Study with id={id} can not be found"}, HTTPStatus.NOT_FOUND

        # validate requst body
        data = request.get_json()
        errors = study_change_org_schema.validate(data)
        if errors:
            return {
                "msg": "Request body is incorrect",
                "errors": errors,
            }, HTTPStatus.BAD_REQUEST

        # get organization which should be deleted
        org_id = data["id"]
        organization = db.Organization.get(org_id)
        if not organization:
            return {
                "msg": f"Organization with id={org_id} is not found"
            }, HTTPStatus.NOT_FOUND

        if not self.r.can_for_col(P.EDIT, study.collaboration_id):
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        # delete organization and update
        study.organizations.remove(organization)
        study.save()
        return org_schema.dump(study.organizations, many=True), HTTPStatus.OK



import logging

from flask import g
from vantage6.server import db

from vantage6.backend.common.permission import RuleCollectionBase, PermissionManagerBase
from vantage6.backend.common.resource.error_handling import UnauthorizedError
from vantage6.server.model.base import Base
from vantage6.server.model.role import Role
from vantage6.server.model.rule import Rule, Operation, Scope
from vantage6.common import logger_name

module_name = logger_name(__name__)
log = logging.getLogger(module_name)


def obtain_auth_collaborations() -> list[db.Collaboration]:
    """
    Obtain the collaborations that the auth is part of.

    Returns
    -------
    list[db.Collaboration]
        List of collaborations
    """
    if g.user:
        return g.user.organization.collaborations
    elif g.node:
        return g.node.organization.collaborations
    else:
        return [db.Collaboration.get(g.container["collaboration_id"])]


def obtain_auth_organization() -> db.Organization:
    """
    Obtain the organization model from the auth that is logged in.

    Returns
    -------
    db.Organization
        Organization model
    """
    if g.user:
        org_id = g.user.organization.id
    elif g.node:
        org_id = g.node.organization.id
    else:
        org_id = g.container["organization_id"]
    return db.Organization.get(org_id)


class RuleCollection(RuleCollectionBase):
    """
    Class that tracks a set of all rules for a certain resource name for
    permissions of the vantage6 server.
    """

    def allowed_for_org(self, operation: Operation, subject_org_id: int | str) -> bool:
        """
        Check if an operation is allowed on a certain organization

        Parameters
        ----------
        operation: Operation
            Operation to check if allowed
        subject_org_id: int | str
            Organization id on which the operation should be allowed. If a
            string is given, it will be converted to an int

        Returns
        -------
        bool
            True if the operation is allowed on the organization, False
            otherwise
        """
        if isinstance(subject_org_id, str):
            subject_org_id = int(subject_org_id)

        auth_org = obtain_auth_organization()

        # check if the entity has global permission
        global_perm = getattr(self, f"{operation}_{Scope.GLOBAL}")
        if global_perm and global_perm.can():
            return True

        # check if the entity has organization permission and organization is
        # the same as the subject organization
        org_perm = getattr(self, f"{operation}_{Scope.ORGANIZATION}")
        if auth_org.id == subject_org_id and org_perm and org_perm.can():
            return True

        # check if the entity has collaboration permission and the subject
        # organization is in the collaboration of the own organization
        col_perm = getattr(self, f"{operation}_{Scope.COLLABORATION}")
        if col_perm and col_perm.can():
            for col in auth_org.collaborations:
                if subject_org_id in [org.id for org in col.organizations]:
                    return True

        # no permission found
        return False

    def can_for_col(self, operation: Operation, collaboration_id: int | str) -> bool:
        """
        Check if the user or node can perform the operation on a certain
        collaboration

        Parameters
        ----------
        operation: Operation
            Operation to check if allowed
        collaboration_id: int | str
            Collaboration id on which the operation should be allowed. If a
            string is given, it will be converted to an int
        """
        if isinstance(collaboration_id, str):
            collaboration_id = int(collaboration_id)

        auth_collabs = obtain_auth_collaborations()

        # check if the entity has global permission
        global_perm = getattr(self, f"{operation}_{Scope.GLOBAL}")
        if global_perm and global_perm.can():
            return True

        # check if the entity has collaboration permission and the subject
        # collaboration is in the collaborations of the user/node
        col_perm = getattr(self, f"{operation}_{Scope.COLLABORATION}")
        if (
            col_perm
            and col_perm.can()
            and self._id_in_list(collaboration_id, auth_collabs)
        ):
            return True

        # no permission found
        return False

    def get_max_scope(self, operation: Operation) -> Scope | None:
        """
        Get the highest scope that the entity has for a certain operation

        Parameters
        ----------
        operation: Operation
            Operation to check

        Returns
        -------
        Scope | None
            Highest scope that the entity has for the operation. None if the
            entity has no permission for the operation
        """
        if getattr(self, f"{operation}_{Scope.GLOBAL}"):
            return Scope.GLOBAL
        elif getattr(self, f"{operation}_{Scope.COLLABORATION}"):
            return Scope.COLLABORATION
        elif getattr(self, f"{operation}_{Scope.ORGANIZATION}"):
            return Scope.ORGANIZATION
        elif getattr(self, f"{operation}_{Scope.OWN}"):
            return Scope.OWN
        else:
            return None

    def has_at_least_scope(self, scope: Scope, operation: Operation) -> bool:
        """
        Check if the entity has at least a certain scope for a certain
        operation

        Parameters
        ----------
        scope: Scope
            Scope to check if the entity has at least
        operation: Operation
            Operation to check

        Returns
        -------
        bool
            True if the entity has at least the scope, False otherwise
        """
        scopes: list[Scope] = self._get_scopes_from(scope)
        for s in scopes:
            perm = getattr(self, f"{operation}_{s}", None)
            if perm and perm.can():
                return True
        return False

    def _id_in_list(self, id_: int, resource_list: list[Base]) -> bool:
        """
        Check if resource list contains a resource with a certain ID

        Parameters
        ----------
        id_ : int
            ID of the resource
        resource_list : list[db.Base]
            List of resources

        Returns
        -------
        bool
            True if resource is in list, False otherwise
        """
        return any(r.id == id_ for r in resource_list)

    def _get_scopes_from(self, minimal_scope: Scope) -> list[Scope]:
        """
        Get scopes that are at least equal to a certain scope

        Parameters
        ----------
        minimal_scope: Scope
            Minimal scope

        Returns
        -------
        list[Scope]
            List of scopes that are at least equal to the minimal scope

        Raises
        ------
        ValueError
            If the minimal scope is not known
        """
        if minimal_scope == Scope.ORGANIZATION:
            return [Scope.ORGANIZATION, Scope.COLLABORATION, Scope.GLOBAL]
        elif minimal_scope == Scope.COLLABORATION:
            return [Scope.COLLABORATION, Scope.GLOBAL]
        elif minimal_scope == Scope.GLOBAL:
            return [Scope.GLOBAL]
        elif minimal_scope == Scope.OWN:
            return [Scope.OWN, Scope.ORGANIZATION, Scope.COLLABORATION, Scope.GLOBAL]
        else:
            raise ValueError(f"Unknown scope '{minimal_scope}'")


class PermissionManager(PermissionManagerBase):
    def assign_rule_to_node(
        self, resource: str, scope: Scope, operation: Operation
    ) -> None:
        """
        Assign a rule to the Node role.

        Parameters
        ----------
        resource: str
            Resource that the rule applies to
        scope: Scope
            Scope that the rule applies to
        operation: Operation
            Operation that the rule applies to
        """
        self.assign_rule_to_fixed_role(
            self.default_roles.NODE, resource, operation, scope
        )

    def assign_rule_to_container(
        self, resource: str, scope: Scope, operation: Operation
    ) -> None:
        """
        Assign a rule to the container role.

        Parameters
        ----------
        resource: str
            Resource that the rule applies to
        scope: Scope
            Scope that the rule applies to
        operation: Operation
            Operation that the rule applies to
        """
        self.assign_rule_to_fixed_role(
            self.default_roles.CONTAINER, resource, operation, scope
        )

    def assign_rule_to_fixed_role(
        self, fixedrole: str, resource: str, operation: Operation, scope: Scope
    ) -> None:
        """
        Attach a rule to a fixed role (not adjustable by users).

        Parameters
        ----------
        fixedrole: str
            Name of the fixed role that the rule should be added to
        resource: str
            Resource that the rule applies to
        scope: Scope
            Scope that the rule applies to
        operation: Operation
            Operation that the rule applies to
        """
        role = Role.get_by_name(fixedrole)
        if not role:
            log.warning(f"{fixedrole} role not found, creating it now!")
            role = Role(
                name=fixedrole, description=f"{fixedrole} role", is_default_role=True
            )

        rule = Rule.get_by_(name=resource, scope=scope, operation=operation)
        rule_params = f"{resource},{scope},{operation}"

        if not rule:
            log.error(f"Rule ({rule_params}) not found!")

        if rule not in role.rules:
            role.rules.append(rule)
            log.info(f"Rule ({rule_params}) added to " f"{fixedrole} role!")

    def get_new_collection(self, name: str) -> RuleCollection:
        """
        Initialize and return a new ServerRuleCollection.
        Parameters
        ----------
        name: str
            Name of the collection

        Returns
        -------
        RuleCollectionBase
            New ServerRuleCollection
        """
        return RuleCollection(name)

    def register_rule(
        self,
        resource: str,
        scope: Scope,
        operation: Operation,
        description=None,
        assign_to_node=False,
        assign_to_container=False,
    ) -> None:
        """
        Register a permission rule in the database with the scope.

        If a rule already exists, nothing is done. This rule can be used in API
        endpoints to determine if a user, node or container can do a certain
        operation in a certain scope.

        Parameters
        ----------
        resource : str
            API resource that the rule applies to
        operation : Operation
            Operation of the rule
        scope : Scope
            Scope of the rule
        description : String, optional
            Human readable description where the rule is used for, by default
            None
        assign_to_node: bool, optional
            Whether rule should be assigned to the node role or not. Default
            False
        assign_to_container: bool, optional
            Whether rule should be assigned to the container role or not.
            Default False

        """

        rule = Rule.get_by_(name=resource, scope=scope, operation=operation)
        if not rule:
            rule = Rule(
                name=resource, operation=operation, scope=scope, description=description
            )
            rule.save()
            log.debug(
                "New auth rule '%s' with scope=%s and operation=%s is stored in the DB",
                resource,
                scope,
                operation,
            )

        if assign_to_container:
            self.assign_rule_to_container(resource, scope, operation)

        if assign_to_node:
            self.assign_rule_to_node(resource, scope, operation)

        # assign all new rules to root user
        self.assign_rule_to_root(resource, operation, scope)

        self.collection(resource).add(rule.operation, rule.scope)

    def check_user_rules(self, rules: list[Rule]) -> None:
        """
        Check if a user, node or container has all the `rules` in a list

        Parameters
        ----------
        rules: list[:class:`~vantage6.server.model.rule.Rule`]
            List of rules that user is checked to have

        Raises
        ---------
        UnauthorizedError
        """
        for rule in rules:
            if not self.collections[rule.name].has_at_least_scope(
                rule.scope, rule.operation
            ):
                raise UnauthorizedError(
                    f"You don't have the rule ({rule.name}, "
                    f"{rule.scope.name.lower()}, "
                    f"{rule.operation.name.lower()})"
                )



import logging

from datetime import datetime, timedelta, timezone

from vantage6.common.task_status import TaskStatus
from vantage6.server.model import Run
from vantage6.server.model.base import DatabaseSessionManager
from vantage6.server.service.azure_storage_service import AzureStorageService

module_name = __name__.split(".")[-1]
log = logging.getLogger(module_name)


def cleanup_runs_data(config: dict, include_input: bool = False):
    """
    Clear the `result` and (optionally) `input` field for `Run` instances older
    than the specified number of days.

    Parameters
    ----------
    days : int
        The number of days after which results should be cleared.
    """
    days = config.get("runs_data_cleanup_days")
    azure_config = config.get("large_result_store", {})
    if azure_config:
        storage_adapter = AzureStorageService(azure_config)
    threshold_date = datetime.now(timezone.utc) - timedelta(days=days)
    session = DatabaseSessionManager.get_session()

    if not days or days < 1:
        log.warning(
            "Invalid number of days '%s' specified. No cleanup will be performed.", days
        )
        return

    try:
        with session.begin():
            runs = (
                session.query(Run)
                .filter(
                    Run.finished_at < threshold_date,
                    Run.cleanup_at == None,
                    Run.status == TaskStatus.COMPLETED,
                )
                .all()
            )

            for run in runs:
                if (
                    run.result is not None
                    and run.blob_storage_used == True
                    and storage_adapter
                ):
                    log.debug(f"Deleting blob: {run.result}")
                    try:
                        storage_adapter.delete_blob(run.result)
                    except Exception as e:
                        log.warning(f"Failed to delete result {run.result}: {e}")
                run.result = ""
                if include_input:
                    if (
                        run.input is not None
                        and run.blob_storage_used == True
                        and storage_adapter
                    ):
                        log.debug(f"Deleting blob: {run.input}")
                        try:
                            storage_adapter.delete_blob(run.input)
                        except Exception as e:
                            log.warning(f"Failed to delete input {run.input}: {e}")
                    run.input = ""
                run.cleanup_at = datetime.now(timezone.utc)
                log.info(f"Cleared result for Run ID {run.id}.")

        log.info(
            "Cleanup job completed successfully, deleted %d old run results.", len(runs)
        )
    except Exception as e:
        log.error(f"Failed to cleanup old run results: {e}")
        raise



"""
Resources below '/<api_base>/version'
"""

import logging

from flask_restful import Api
from http import HTTPStatus
from vantage6.common import logger_name
from vantage6.server.resource import ServicesResources
from vantage6.server._version import __version__


module_name = logger_name(__name__)
log = logging.getLogger(module_name)


def setup(api: Api, api_base: str, services: dict) -> None:
    """
    Setup the version resource.

    Parameters
    ----------
    api : Api
        Flask restful api instance
    api_base : str
        Base url of the api
    services : dict
        Dictionary with services required for the resource endpoints
    """
    path = "/".join([api_base, module_name])
    log.info(f'Setting up "{path}" and subdirectories')

    api.add_resource(
        Version,
        path,
        endpoint="version",
        methods=("GET",),
        resource_class_kwargs=services,
    )


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
class Version(ServicesResources):
    def get(self):
        """Get version
        ---
        description: Return the version of the server instance

        responses:
          200:
            description: Ok

        tags: ["Other"]
        """
        return {"version": __version__}, HTTPStatus.OK



import os
import logging
import requests
from flask import Response, request
from http import HTTPStatus

from vantage6.common.enum import AlgorithmViewPolicies, StorePolicies
from vantage6.backend.common.globals import HOST_URI_ENV
from vantage6.backend.common import get_server_url
from vantage6.server import db


module_name = __name__.split(".")[-1]
log = logging.getLogger(module_name)


def post_algorithm_store(
    data: dict, config: dict, headers: dict | None = None
) -> tuple[dict | db.AlgorithmStore, HTTPStatus]:
    """Add algorithm store to a collaboration

    Parameters
    ----------
    data : dict
        Request body as required for POST /algorithmstore request
    config : dict
        Server configuration
    headers : dict | None
        Headers to be included in the request. Usually, these will be Authorization
        headers

    Returns
    -------
    tuple[dict | db.AlgorithmStore, HTTPStatus]
        Response or AlgorithmStore object and HTTP status code
    """
    # check if algorithm store is already available for the collaboration
    collaboration_id = data.get("collaboration_id", None)
    algorithm_store_url = data["algorithm_store_url"]
    if algorithm_store_url.endswith("/"):
        algorithm_store_url = algorithm_store_url[:-1]
    existing_algorithm_stores = db.AlgorithmStore.get_by_url(algorithm_store_url)
    records_to_delete = []
    if existing_algorithm_stores:
        collabs_with_algo_store = [
            a.collaboration_id for a in existing_algorithm_stores
        ]
        if None in collabs_with_algo_store:
            return {
                "msg": "Algorithm store is already available for all collaborations"
            }, HTTPStatus.BAD_REQUEST
        if collaboration_id in collabs_with_algo_store:
            return {
                "msg": "Algorithm store is already available for this collaboration"
            }, HTTPStatus.BAD_REQUEST
        if not collaboration_id:
            # algorithm store is currently available for some
            # collaborations, but now it will be available for all of them.
            # Remove the records that only make it available to some
            # collaborations (this prevents duplicates)
            records_to_delete = existing_algorithm_stores

    # raise a warning if the algorithm store url is insecure (i.e. localhost)
    force = data.get("force", False)
    if not force and (
        "localhost" in algorithm_store_url or "127.0.0.1" in algorithm_store_url
    ):
        return {
            "msg": "Algorithm store url is insecure: localhost services "
            "may be run on any computer. Add it anyway by setting the "
            "'force' flag to true, but only do so for development servers!"
        }, HTTPStatus.BAD_REQUEST

    server_url = get_server_url(config, data.get("server_url"))
    if not server_url:
        return {
            "msg": "The 'server_url' key is required in the server "
            "configuration, or as a parameter. Please add it or ask your "
            "server administrator to specify it in the server configuration."
        }, HTTPStatus.BAD_REQUEST

    # whitelist this vantage6 server url for the algorithm store
    response, status = request_algo_store(
        algo_store_url=algorithm_store_url,
        server_url=server_url,
        endpoint="vantage6-server",
        method="post",
        force=force,
        headers=headers,
        params={"url": server_url},
    )
    if status == HTTPStatus.FORBIDDEN:
        # if whitelisting of the server at the algorithm store fails with status
        # forbidden, the store does not allow the server to be whitelisted. Check if
        # the store has open algorithms. If it does, the server can still whitelist the
        # store to get the open algorithms, but the users of this server will not be
        # able to manage anything at the store.
        if store_has_open_algorithms(algorithm_store_url, server_url):
            log.warning(
                "Could not whitelist the current server at algorithm store %s."
                " This server will only have read-only access to algorithms "
                "from the store",
                algorithm_store_url,
            )
        else:
            # if the store does not have open algorithms, the server should not
            # whitelist it as it will not be able to do anything with the store - return
            # the error message from the algorithm store
            return response, status
    elif status not in [HTTPStatus.CREATED, HTTPStatus.ALREADY_REPORTED]:
        # return error
        return response, status

    # delete and create records
    for record in records_to_delete:
        record.delete()
    algorithm_store = db.AlgorithmStore(
        name=data["name"],
        url=algorithm_store_url,
        collaboration_id=collaboration_id,
    )
    algorithm_store.save()

    return algorithm_store, HTTPStatus.CREATED


def store_has_open_algorithms(algorithm_store_url: str, server_url: str) -> bool:
    """Check if the algorithm store has open algorithms

    Parameters
    ----------
    algorithm_store_url : str
        URL to the algorithm store
    server_url: str
        URL to the current vantage6 server

    Returns
    -------
    bool
        True if the store has open algorithms, False otherwise
    """
    response, status_code = request_algo_store(
        algo_store_url=algorithm_store_url,
        server_url=server_url,
        endpoint="/policy/public",
        method="get",
    )
    if status_code != HTTPStatus.OK:
        return False
    try:
        return (
            response.json().get(StorePolicies.ALGORITHM_VIEW, None)
            == AlgorithmViewPolicies.PUBLIC
        )
    except KeyError:
        return False


def request_algo_store(
    algo_store_url: str,
    server_url: str,
    endpoint: str,
    method: str,
    force: bool = False,
    headers: dict = None,
    params: dict = None,
) -> tuple[dict | Response, HTTPStatus]:
    """
    Whitelist this vantage6 server url for the algorithm store.

    Parameters
    ----------
    algo_store_url : str
        URL to the algorithm store
    server_url : str
        URL to this vantage6 server. This is used to whitelist this server
        at the algorithm store.
    endpoint : str
        Endpoint to use at the algorithm store.
    method : str
        HTTP method to use.
    force : bool
        If True, the algorithm store will be added even if the algorithm
        store url is insecure (i.e. localhost)
    headers : dict
        Headers to be included in the request. Usually, these will be Authorization
        headers

    Returns
    -------
    tuple[dict | Response, HTTPStatus]
        The response of the algorithm store and the HTTP status. If the
        request to the algorithm store is unsuccessful, a dict with an error message is
        returned instead of the response.
    """
    is_localhost_algo_store = _contains_localhost(algo_store_url)
    try:
        response = _execute_algo_store_request(
            algo_store_url, server_url, endpoint, method, force, headers, params
        )
    except requests.exceptions.ConnectionError as exc:
        if not is_localhost_algo_store:
            log.warning("Request to algorithm store failed")
            log.exception(exc)
        response = None

    if not response and is_localhost_algo_store:
        # try again with the docker host ip
        host_uri = os.environ.get(HOST_URI_ENV, None)
        if not host_uri:
            msg = (
                "You are trying to connect to a localhost algorithm store, but this "
                "refers to the container itself. Please set the configuration option "
                "'host_uri' in the 'dev' section  of the config file to the host's IP "
                "address."
            )
            log.warning(msg)
            return {"msg": msg}, HTTPStatus.BAD_REQUEST
        algo_store_url = algo_store_url.replace("localhost", host_uri).replace(
            "127.0.0.1", host_uri
        )
        # replace double http:// with single
        algo_store_url = algo_store_url.replace("http://http://", "http://")
        try:
            response = _execute_algo_store_request(
                algo_store_url, server_url, endpoint, method, force, headers, params
            )
        except requests.exceptions.ConnectionError as exc:
            log.warning("Request to algorithm store failed")
            log.exception(exc)
            response = None

    if response is None:
        return {
            "msg": "Algorithm store cannot be reached. Make sure that "
            "it is online and that you have included the API path (default /api) at the"
            " end of the algorithm store URL"
        }, HTTPStatus.NOT_FOUND
    elif response.status_code not in [HTTPStatus.CREATED, HTTPStatus.OK]:
        try:
            msg = (
                f"Algorithm store error: {response.json()['msg']}, HTTP status: "
                f"{response.status_code}"
            )
        except (KeyError, requests.exceptions.JSONDecodeError):
            msg = (
                "Communication to algorithm store failed. HTTP status: "
                f"{response.status_code}"
            )
        return {"msg": msg}, response.status_code
    # else: server has been registered at algorithm store, proceed
    return response, response.status_code


# TODO move this to a utility module - it is also in algo store code
def _contains_localhost(url: str) -> bool:
    """Check if the url refers to localhost address"""
    return url.startswith("http://localhost") or url.startswith("http://127.0.0.1")


def _execute_algo_store_request(
    algo_store_url: str,
    server_url: str,
    endpoint: str,
    method: str,
    force: bool,
    headers: dict = None,
    param_dict: dict = None,
) -> requests.Response:
    """
    Send a request to the algorithm store to whitelist this vantage6 server
    url for the algorithm store.

    Parameters
    ----------
    algo_store_url : str
        URL to the algorithm store
    server_url : str
        URL to this vantage6 server. This is used to whitelist this server
        at the algorithm store.
    endpoint : str
        Endpoint to use at the algorithm store.
    method : str
        HTTP method to use. Choose "post" for adding the server url and
        "delete" for removing it.
    force : bool
        If True, the algorithm store will be added even if the algorithm
        store url is insecure (i.e. localhost)
    headers : dict
        Headers to be included in the request. Usually, these will be Authorization
        headers
    params : dict
        Parameters to be included in the request

    Returns
    -------
    requests.Response | None
        Response from the algorithm store. If the algorithm store is not
        reachable, None is returned
    """
    if server_url.endswith("/"):
        server_url = server_url[:-1]
    if algo_store_url.endswith("/"):
        algo_store_url = algo_store_url[:-1]

    param_dict = param_dict if param_dict is not None else {}
    if force:
        param_dict["force"] = True

    # set headers
    if not headers:
        headers = {}
    headers["server_url"] = server_url
    if "Authorization" in request.headers and not headers.get("Authorization"):
        headers["Authorization"] = request.headers["Authorization"]

    params = None
    json = None
    method = method.lower()
    if method == "get":
        request_function = requests.get
        params = param_dict
    elif method == "post":
        request_function = requests.post
        json = param_dict
    elif method == "delete":
        request_function = requests.delete
        params = param_dict
    else:
        raise ValueError(f"Method {method} not supported")

    return request_function(
        f"{algo_store_url}/{endpoint}",
        params=params,
        json=json,
        headers=headers,
    )
